/********************************************************************
 Mod5282 time setting example program (requires Carrier board)
 This example program will get and set the time for use with the
 time() function in 3 ways:
 1. Using a network time server (NTP)
 2. Using the Real Time Clock on the Mod5282 Carrier development board
 3. Manually

 Revision 1.0, 6/10/2003
 
 Requires revision 1.80 tools plus ioboard update
 
*********************************************************************/ 
#include "predef.h" 
#include <stdio.h>
#include <ctype.h>
#include <startnet.h>
#include <autoupdate.h>
#include <dhcpclient.h>
#include <time.h>
#include <udp.h>
#include <tcp.h>
#include <string.h>

#include "..\..\mod5282\system\ioboardrtc.h"


//----- function prototypes -----
extern "C" { 
	void UserMain(void * pd);
}

time_t set_time(time_t time_to_set);


/**************** NTP example code to set the clock  ****************
 The first IP address in this list is a netburner time server. 
 The second address is a public NTP server, it is provided by the 
 government and has specific rules for usage. 
 See: http://www.eecis.udel.edu/~mills/ntp/servers.html 
 and search the page for nist1.aol-ca.truetime.com for details. 
 If your local network has an NTP server, it is strongly recommended 
 that you use it.
********************************************************************/
const char * NTPServers[] = {"66.80.176.21","207.200.81.113","\0"};

/* The structure to hold the NTP packet */
struct NTP_Packet
{
	BYTE LV_VN_Mode;
	BYTE Stratum;
	BYTE Poll;
	signed char Precision;
	
	DWORD Root_Delay;
	DWORD Root_Dispersion;
	BYTE  Ref_Id[4];
	
	DWORD Ref_Time_Secs;
	DWORD Ref_Time_Frac;
	
	DWORD Org_Time_Secs;
	DWORD Org_Time_Frac;
	
	DWORD Rx_Time_Secs;
	DWORD Rx_Time_Frac;
	
	DWORD Tx_Time_Secs;
	DWORD Tx_Time_Frac;
};



/* Functions to manipulate a 64 bit long long */
unsigned long long MakeLongLong(DWORD msb, DWORD lsb)
{
	unsigned long long lv;
	PDWORD pdw=(PDWORD)&lv;
	pdw[0]=msb;
	pdw[1]=lsb;
	return lv;
}

/* Get the leastsignificat DWORD of a long long  usually the 
   seconds fraction in an NTP long long 
*/
DWORD GetLsb(unsigned long long lv)
{
	return *(((PDWORD)&lv)+1);
}

/* Get the most significant DWORD of a long long  usually the 
   seconds in an NTP long long 
*/
DWORD GetMsb(unsigned long long lv)
{
	 return *(((PDWORD)&lv));
}

#define NTPRXPORT (1500) /* This value is completly arbitrary */
#define NTPTXPORT (123)  /* This value is set by the standard */



/* This function will try to get time from an NTP server.
   It will return the NTP time to the nearest second if successful,
   or return 0 if it fails.
*/
DWORD GetNTPTime(IPADDR NTP_server_ip)
{
	OS_FIFO fifo;
	OSFifoInit(&fifo);
	
	//Register to listen for UDP packets on port number 'port'
	RegisterUDPFifo(NTPRXPORT, &fifo);
	UDPPacket upkt;
	
	NTP_Packet * pNtp=(NTP_Packet *) upkt.GetDataBuffer();
	upkt.SetDataSize(sizeof(NTP_Packet)); 
	
	bzero((char *)pNtp,sizeof(NTP_Packet));
	
	pNtp->LV_VN_Mode=0xB;
	
	pNtp->Ref_Time_Secs=0;
	pNtp->Ref_Time_Frac=0;
	
	pNtp->Org_Time_Secs=0;
	pNtp->Org_Time_Frac=0;
	
	upkt.SetSourcePort(NTPRXPORT); /* NTP will return the result to the port it came from */
	upkt.SetDestinationPort(NTPTXPORT);
	
	DWORD lcl_tx_time=TimeTick;
	
	upkt.Send(NTP_server_ip);
	
	UDPPacket urpkt(&fifo,10*TICKS_PER_SECOND);
	
	DWORD lcl_rx_time=TimeTick;
	
	UnregisterUDPFifo(1500); 
	
	 //Did we get a valid packet, or just time out?
	 if (urpkt.Validate())
	  {
	   WORD len=urpkt.GetDataSize();
	   if (len>=sizeof(NTP_Packet ))
		{
		 NTP_Packet * prNtp=(NTP_Packet *) urpkt.GetDataBuffer();
		 
		 unsigned long long rmt_rx_tim= MakeLongLong(prNtp->Rx_Time_Secs,prNtp->Rx_Time_Frac); 
		 unsigned long long rmt_tx_tim= MakeLongLong(prNtp->Tx_Time_Secs,prNtp->Tx_Time_Frac); 
		 unsigned long long dt=rmt_tx_tim-rmt_rx_tim;
	
		 dt/=2;
	
		 rmt_rx_tim+=dt; /* Remote center time */
		 /* Now to adjust for the timing round trip */
		 
		 DWORD tick_at_time = (lcl_tx_time + lcl_rx_time) / 2;
		 /* So when our time tick was tick_at_time the real time was rmt_rx_tim */
	
		 DWORD tick_adjust=TimeTick-tick_at_time;
		 /* Now we need to add these ticks  to the remote time */
		 /* With the standard NB 20 ticks per second  */
		 unsigned long long adjust=tick_adjust;
		 adjust *= 214748364; /* Each local tick is 214784363 NTP counts */
		 rmt_rx_tim += adjust;
		 return GetMsb(rmt_tx_tim);
		}
	  }
	 return 0;
}
#define NTP_TimeOffset ((DWORD)(2208988800U))


/*********************************************************************
 This function will set the clock using NTP. It returns non zero 
 if succesful 
*********************************************************************/
DWORD RobustSetTime()
{
	int index=0;
	do
	{
		DWORD dwv = GetNTPTime(AsciiToIp(NTPServers[index]));
		if (dwv > 0 ) 
		{
			//iprintf("Got NTP time from %s\r\n",NTPServers[index]);
			 dwv=dwv-NTP_TimeOffset;
			 set_time(dwv); 
			 return dwv;
		}
	
		int fd=connect(AsciiToIp(NTPServers[index]),0,37,20);
		if (fd >0)
		{
			dwv=0;
			int rv=read(fd,(char *)&dwv,4);
			close(fd);
			if (rv==4) 
			{
				//iprintf("Got TCP DayTime time from %s\r\n",NTPServers[index]);
				dwv=dwv-NTP_TimeOffset;
				set_time(dwv); 
				return dwv;
			}
		 }
		index++;
	} while (NTPServers[index][0]);
	return 0;
}



/********************************************************************
 UserMain
********************************************************************/ 
void UserMain(void * pd)
{
	InitializeStack();
	if (EthernetIP == 0)
		GetDHCPAddress(); 
	OSChangePrio(MAIN_PRIO);
	EnableAutoUpdate();
	
	iprintf("Application started\n");
	while (1)
	{
		iprintf("Select option to set system time:\r\n");
		iprintf("N - Get NTP time and set both the system and RTC clocks\r\n");
		iprintf("S - Manually set both the system and RTC time\r\n");
		iprintf("G - Get the time from the RTC and set the system time\r\n");
		char c = getchar();
		switch (toupper(c))
		{
			case 'N':
			{
				iprintf("Getting time from NTP...\r\n");
				if (RobustSetTime()) 
					iprintf("Time set\r\n");
				else
					iprintf("Failed to set time\r\n");

				if(IOBoardRTCSetRTCfromSystemTime() == 0)
					iprintf("RTC Time is now set\r\n");
				else
					iprintf("RTC failed to set \r\n");
			}
			break;

			case 'S':
			{
				int clock_offset;
				struct tm bts;
				bts.tm_isdst = 0;
				
				int i;
				iprintf("Clock offset from GMT (in hours)?)");
				scanf("%d",&clock_offset);
				
				iprintf("\r\nYear yyyy?)");
				scanf("%d",&i);
				bts.tm_year=i-1900;
				
				iprintf("\r\nMonth mm (1..12 etc)?");
				scanf("%d",&i);
				bts.tm_mon=(BYTE)i-1;
				
				iprintf("\r\nDay of Month dd (1..31 etc)?");
				scanf("%d",&i);
				bts.tm_mday=(BYTE)i;
				
				iprintf("\r\nHour hh (0..23)?");
				scanf("%d",&i);
				bts.tm_hour=i;
				
				iprintf("\r\nMinute mm (0..59)?");
				scanf("%d",&i);
				bts.tm_min=i;
				
				iprintf("\r\nSeconds ss (0..59)?");
				scanf("%d",&i);
				bts.tm_sec=i;
				
				time_t t=mktime(&bts);
				iprintf("Local Time = %s",ctime(&t)); 
				
				t-=(clock_offset*3600);
				
				iprintf("GMT Time = %s",ctime(&t)); 
				set_time(t); 

				if(IOBoardRTCSetRTCfromSystemTime()==0)
					iprintf("RTC time is now set\r\n");
				else
					iprintf("RTC failed to set \r\n");
				}
				break;

				case 'G':
				{
					iprintf("Getting time from RTC on I/O board\r\n");
					if(IOBoardRTCSetSystemFromRTCTime() == 0)
						iprintf("System Time is now set\r\n");
					else
						iprintf("Failed to get time from RTC\r\n");
				}
				break;
		   }

		while (1)
		{
			time_t time_val = time(NULL);
			iprintf("GMT Time = %s",ctime(&time_val)); 
			time_val-=(7*3600); /* 7 hours for local time offset to PDT */
			iprintf("Local Time = %s",ctime(&time_val)); 
			OSTimeDly(20);
		}
	}
}
